from itertools import chain, groupby
from typing import Any, Dict, Iterator, List, Mapping, Optional, Set, Tuple, Union

from attr import attrib, dataclass

from .ast_node import (
    Background,
    Comment,
    DataTable,
    DocString,
    Examples,
    Feature,
    GherkinDocument,
    Location,
    Node,
    Scenario,
    ScenarioOutline,
    Step,
    TableRow,
    Tag,
    TagGroup,
)
from .options import AlignmentMode, TagLineMode
from .utils import camel_to_snake_case, extract_beginning_spaces, get_display_width

INDENT_LEVEL_MAP: Mapping[Any, int] = {
    Feature: 0,
    Background: 1,
    Scenario: 1,
    ScenarioOutline: 1,
    Step: 2,
    Examples: 2,
    TableRow: 3,
}


def generate_language_header(language: str) -> Comment:
    return Comment(Location(1, 1), f"# language: {language}")  # type: ignore


def generate_step_line(
    step: Step,
    keyword_alignment: AlignmentMode,
    indent: str,
    *,
    keyword_padding_width: int = 0,
) -> str:
    """
    Generate lines for steps. The step keywords are aligned according to the parameter
    `keyword_alignment`. For example:

    If `keyword_alignment = AlignmentMode.NONE`:
        Given Enter search term 'Cucumber'
        When Do search
        Then Single result is shown for 'Cucumber'

    If `keyword_alignment = AlignmentMode.LEFT`:
        Given Enter search term 'Cucumber'
        When  Do search
        Then  Single result is shown for 'Cucumber'

    If `keyword_alignment = AlignmentMode.Right`:
        Given Enter search term 'Cucumber'
         When Do search
         Then Single result is shown for 'Cucumber'
    """
    indent_level: int = INDENT_LEVEL_MAP[Step]

    formatted_keyword = format_step_keyword(
        step.keyword, keyword_alignment, keyword_padding_width=keyword_padding_width
    )

    return f"{indent * indent_level}{formatted_keyword} {step.text}"


def format_step_keyword(
    keyword: str, keyword_alignment: AlignmentMode, *, keyword_padding_width: int = 0
) -> str:
    """
    Insert padding to step keyword if necessary based on how we want to align them.
    """
    if keyword_alignment is AlignmentMode.NONE or keyword_padding_width <= 0:
        return keyword

    padding = " " * (keyword_padding_width - get_display_width(keyword))

    if keyword_alignment is AlignmentMode.LEFT:
        return keyword + padding
    else:
        return padding + keyword


def generate_keyword_line(
    keyword: str, name: str, indent: str, indent_level: int
) -> str:
    return f"{indent * indent_level}{keyword}: {name}".rstrip()


def generate_description_lines(
    description: Optional[str], indent: str, indent_level: int
) -> List[str]:
    description_lines = extract_description_lines(description)

    lines = [f"{indent * indent_level}{line}" for line in description_lines]

    # Add an empty line after the description, if it exists
    if lines:
        lines.append("")

    return lines


def extract_description_lines(description: Optional[str]) -> List[str]:
    if description is None:
        return []

    return description.splitlines()


def generate_table_lines(rows: List[TableRow], indent: str) -> List[str]:
    """
    Generate lines for table. The columns in a table need to have the same width.
    """
    if not rows:
        return []

    indent_level = INDENT_LEVEL_MAP[TableRow]

    n_columns = len(rows[0])

    # Find the max width of a cell in a column, so that every cell in the same column
    # has the same width
    column_widths = [
        max(get_display_width(row[column_index].value) for row in rows)
        for column_index in range(n_columns)
    ]

    lines = []
    for row in rows:
        line = "|"

        for column_index in range(n_columns):
            # Left-align the content of each cell, fix the width of the cell
            content = row[column_index].value
            column_width = column_widths[column_index]
            content_width = get_display_width(content)
            padding = " " * (column_width - content_width)
            line += f" {content}{padding} |"

        lines.append(line)

    return [f"{indent * indent_level}{line}" for line in lines]


def extract_rows(node: Union[DataTable, Examples]) -> List[TableRow]:
    """
    Extract table rows from either a Datable or Example instance.
    """
    if isinstance(node, DataTable):
        return list(node.rows)

    rows = []

    if isinstance(node, Examples):
        header = node.table_header
        body = node.table_body

        if header is not None:
            rows.append(header)

        if body is not None:
            rows.extend(body)

    return rows


def generate_doc_string_lines(docstring: DocString, indent: str) -> List[str]:
    raw_lines = docstring.content.splitlines()
    raw_lines = ['"""'] + raw_lines + ['"""']

    indent_level = INDENT_LEVEL_MAP[Step]

    return [f"{indent * indent_level}{line}" for line in raw_lines]


ContextMap = Dict[Union[Comment, Tag, TagGroup, TableRow], Any]
Lines = Iterator[str]


@dataclass
class LineGenerator:
    ast: GherkinDocument
    step_keyword_alignment: AlignmentMode
    tag_line_mode: TagLineMode
    indent: str
    __nodes: List[Node] = attrib(init=False)
    __contexts: ContextMap = attrib(init=False)
    __nodes_with_newline: Set[Node] = attrib(init=False)
    __max_step_keyword_width: int = attrib(init=False)

    def __attrs_post_init__(self):
        # Use `__attrs_post_init__` instead of `property` to avoid re-computing attributes

        self.__nodes = list(self.ast)

        if self.tag_line_mode is TagLineMode.SINGLELINE:
            self.__group_tags()

        self.__nodes.sort(key=lambda node: node.location)

        self.__contexts = self.__construct_contexts()
        self.__nodes_with_newline = self.__find_nodes_with_newline()
        self.__max_step_keyword_width = self.__find_max_step_keyword_width()
        self.__add_language_header()

    def __group_tags(self):
        """
        Group the tags of a node, so that we can render them on a single line.
        """

        tag_groups: List[TagGroup] = []
        node: Node
        for node in self.ast:
            if hasattr(node, "tags"):
                tags: Tuple[Tag, ...] = node.tags

                if tags:
                    # The tag group should be placed at the position of the last tag
                    tag_group = TagGroup(
                        members=tags, context=node, location=tags[-1].location
                    )
                    tag_groups.append(tag_group)

        # After grouping the tags, we need to include the tag groups into
        # the list of nodes and remove the tags from the list.
        self.__nodes = [
            node for node in self.__nodes if not isinstance(node, Tag)
        ] + tag_groups

    def __construct_contexts(self) -> ContextMap:
        """
        Construct the information about the context a certain line might need to know to
        properly format these lines.
        """
        contexts: ContextMap = {}
        nodes = self.__nodes

        for node in nodes:
            if hasattr(node, "context"):
                contexts[node] = node.context  # type: ignore

            # We want tags to have the same indentation level with their parents
            for tag in getattr(node, "tags", []):
                contexts[tag] = node

            if isinstance(node, (DataTable, Examples)):
                # We need to know all rows in a table, so that the columns can be padded
                # to have the same widths across all rows. The context of a row is its
                # reformatted line.
                rows = extract_rows(node)
                lines = generate_table_lines(rows, self.indent)

                for row, line in zip(rows, lines):
                    contexts[row] = line

        contexts.update(self.__construct_contexts_for_comments(nodes))

        return contexts

    @staticmethod
    def __construct_contexts_for_comments(nodes: List[Node]) -> ContextMap:
        # The context of each comment line is the next non-comment line.
        #
        # The steps of the algorithm:
        # 1. Group the nodes into comments and non-comments
        # 2. The first node in each group of non-comments is the context of every node
        #    in the previous group, which consists of comments only.
        #
        # We start with a context of None, this lets us know if the document ends
        # with a block of comments.
        #
        # In the original algorithm, we simply set the context of each comment line
        # to be the next line. This leads to a RecursionError if there are too many
        # consecutive comments.

        contexts: ContextMap = {}
        current_context = None

        groups = groupby(reversed(nodes), lambda n: isinstance(n, Comment))

        for key, group in groups:
            if key is False:
                # The current group consists of non-comments, we set the current context
                # to be the last node in the group, since we grouped in the reverse order
                current_context = list(group)[-1]
            else:
                # The current group consists of comments. These comments should have the
                # same indent level, which is the indent level of the current context.
                for node in group:
                    contexts[node] = current_context  # type: ignore

        return contexts

    def __find_nodes_with_newline(self) -> Set[Node]:
        """
        Find all nodes in the AST that needs a new line after it.
        """
        nodes_with_newline: Set[Node] = set()

        node: Optional[Node] = None

        for node in self.__nodes:
            # We want to add a newline after the Feature line, even if it does not
            # have a description. If the feature has a description, we already add
            # a newline after each description.
            if isinstance(node, Feature) and node.description is None:
                nodes_with_newline.add(node)

            children: List[Node] = []

            # Add an empty line after the last step, including its argument, if any
            if isinstance(node, (Background, Scenario, ScenarioOutline)):
                children = list(chain.from_iterable(node.steps))

            # Add an empty line after an examples table
            if isinstance(node, Examples):
                children = list(node)

            if children:
                last_child = children[-1]
                nodes_with_newline.add(last_child)

        # Add the last node in the AST so that we have an empty line at the end
        if node is not None:
            nodes_with_newline.add(node)

        return nodes_with_newline

    def __find_max_step_keyword_width(self) -> int:
        """
        Find the length of the longest step keyword in the document. This is
        used for aligning step keywords.
        """
        if self.step_keyword_alignment is AlignmentMode.NONE:
            # We don't need to align step keywords in this case.
            return 0

        step_keyword_widths = [
            get_display_width(node.keyword.strip())
            for node in self.ast
            if isinstance(node, Step)
        ]
        if not step_keyword_widths:
            return 0

        return max(step_keyword_widths)

    def __add_language_header(self) -> None:
        """Add a language header if the Feature language is not English."""
        # Exit if the language is English or if there is no Feature node
        feature = self.ast.feature
        if not feature:
            return
        language = feature.language
        if language == "en":
            return

        # Register the language header
        language_header = generate_language_header(language)
        self.__nodes.insert(0, language_header)
        self.__nodes_with_newline.add(language_header)
        self.__contexts[language_header] = self.ast.feature

    def generate(self) -> Lines:
        for node in self.__nodes:
            yield from self.visit(node)

            if node in self.__nodes_with_newline:
                yield ""

    def visit(self, node: Node) -> Lines:
        class_name = type(node).__name__

        yield from getattr(
            self, f"visit_{camel_to_snake_case(class_name)}", self.visit_default
        )(node)

    def visit_default(self, node: Node) -> Lines:
        indent_level = INDENT_LEVEL_MAP.get(type(node), 0)

        if hasattr(node, "keyword") and hasattr(node, "name"):
            yield generate_keyword_line(
                node.keyword, node.name, self.indent, indent_level  # type: ignore
            )

        if hasattr(node, "description"):
            yield from generate_description_lines(
                node.description, self.indent, indent_level + 1  # type: ignore
            )

    def visit_step(self, step: Step) -> Lines:
        yield generate_step_line(
            step,
            self.step_keyword_alignment,
            self.indent,
            keyword_padding_width=self.__max_step_keyword_width,
        )

    def visit_tag(self, tag: Tag) -> Lines:
        context = self.__contexts[tag]

        # Every node type containing tags is included in the indent map, so we don't
        # have to worry about KeyError here
        indent_level = INDENT_LEVEL_MAP[type(context)]

        yield f"{self.indent * indent_level}{tag.name}"

    def visit_tag_group(self, tag_group: TagGroup) -> Lines:
        context = self.__contexts[tag_group]

        indent_level = INDENT_LEVEL_MAP[type(context)]

        line_content = " ".join(tag.name for tag in tag_group.members)

        yield f"{self.indent * indent_level}{line_content}"

    def visit_table_row(self, row: TableRow) -> Lines:
        context = self.__contexts[row]

        yield context

    def visit_comment(self, comment: Comment) -> Lines:
        context = self.__contexts[comment]

        # Find the indent level of this comment line
        if context is None:
            # In this case, this comment line is the last line of the document
            indent_level: Optional[int] = 0
        else:
            # Try to look for the indent level of the context in the mapping. If not
            # successful, then we use the same amount of white spaces to indent as
            # the next line.
            indent_level = INDENT_LEVEL_MAP.get(type(context))

        if indent_level is None:
            next_line = next(self.visit(context))
            indent = extract_beginning_spaces(next_line)
        else:
            indent = self.indent * indent_level

        yield f"{indent}{comment.text}"

    def visit_doc_string(self, docstring: DocString) -> Lines:
        yield from generate_doc_string_lines(docstring, self.indent)



from pathlib import Path
from typing import Iterable, Optional

import click
import yaml

CONFIG_FILE = ".reformat-gherkin.yaml"
SYSTEM_ROOT = Path("/").resolve()


def find_project_root(srcs: Iterable[str]) -> Path:
    """
    Return a directory containing .git, .hg, or .reformat-gherkin.yaml.

    That directory can be one of the directories passed in `srcs` or their
    common parent.

    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.
    """
    if not srcs:
        return SYSTEM_ROOT

    common_base = min(Path(src).resolve() for src in srcs)
    if common_base.is_dir():
        # Append a dummy file so `parents` below returns `common_base_dir`, too.
        common_base /= "dummy-file"

    directory = SYSTEM_ROOT
    for directory in common_base.parents:
        if (directory / ".git").is_dir():
            return directory

        if (directory / ".hg").is_dir():
            return directory

        if (directory / CONFIG_FILE).is_file():
            return directory

    return directory  # pragma: no cover


def read_config_file(
    ctx: click.Context, _: click.Parameter, value: Optional[str]
) -> Optional[str]:
    """
    Inject the configuration from ".reformat-gherkin.yaml" into defaults in `ctx`.

    Returns the path to a successfully found and read configuration file, None
    otherwise.
    """
    if not value:
        root = find_project_root(ctx.params.get("src", ()))
        path = root / CONFIG_FILE
        if path.is_file():
            value = str(path.resolve())
        else:
            return None

    try:
        with open(value, "r") as f:
            config = yaml.safe_load(f)
    except (yaml.YAMLError, OSError) as e:
        raise click.FileError(
            filename=value, hint=f"Error reading configuration file: {e}"
        )

    if not config:
        return None

    if ctx.default_map is None:
        ctx.default_map = {}

    ctx.default_map.update(  # type: ignore  # bad types in .pyi
        {k.replace("--", "").replace("-", "_"): v for k, v in config.items()}
    )
    return value



from enum import Enum, unique
from typing import Optional

from attr import dataclass


@unique
class WriteBackMode(Enum):
    INPLACE = "inplace"
    CHECK = "check"

    @classmethod
    def from_configuration(cls, check: bool) -> "WriteBackMode":
        return WriteBackMode.CHECK if check else WriteBackMode.INPLACE


@unique
class AlignmentMode(Enum):
    NONE = None
    LEFT = "left"
    RIGHT = "right"

    @classmethod
    def from_configuration(cls, alignment: Optional[str]) -> "AlignmentMode":
        return AlignmentMode(alignment)


@unique
class NewlineMode(Enum):
    KEEP = None
    LF = "LF"
    CRLF = "CRLF"

    @classmethod
    def from_configuration(cls, newline: Optional[str]) -> "NewlineMode":
        return NewlineMode(newline)


@unique
class TagLineMode(Enum):
    SINGLELINE = "singleline"
    MULTILINE = "multiline"

    @classmethod
    def from_configuration(cls, single_line_tags: bool) -> "TagLineMode":
        return TagLineMode.SINGLELINE if single_line_tags else TagLineMode.MULTILINE


def get_indent_from_configuration(tab_width: int, use_tabs: bool):
    return "\t" if use_tabs else " " * tab_width


@dataclass(frozen=True)
class Options:
    write_back: WriteBackMode
    step_keyword_alignment: AlignmentMode
    newline: NewlineMode
    tag_line_mode: TagLineMode
    fast: bool
    indent: str



from attr import attrib

from ._base import prepare
from .location import LocationMixin


def normalize_comment_text(text: str) -> str:
    """
    Return a consistently formatted comment from the given Comment instance.
    All comments should have a single space between the hash sign and the content.
    """
    # A comment always start with a hash sign
    normalized_text = text[1:].strip()

    return ("# " + normalized_text).strip()


@prepare
class Comment(LocationMixin):
    text: str = attrib(converter=normalize_comment_text)



from itertools import chain
from typing import Optional, Tuple

from ._base import prepare
from .location import LocationMixin
from .step import Step


@prepare
class Background(LocationMixin):
    keyword: str
    name: str
    steps: Tuple[Step, ...]
    description: Optional[str] = None

    def __iter__(self):
        yield self

        yield from chain.from_iterable(self.steps)



from attr import attrib

from ._base import prepare
from .location import LocationMixin


def escape_table_cell_value(text: str) -> str:
    """
    Escape pipe characters `|` in a table cell's value. Since the pipe characters
    are used to separate cells in a row, we need to replace them by `\\|`.
    """

    return text.replace("|", "\\|")


@prepare
class TableCell(LocationMixin):
    value: str = attrib(converter=escape_table_cell_value)



import io
import textwrap
from typing import Any, Dict, Type, TypeVar

from cattr.converters import Converter
from gherkin.errors import ParserError
from gherkin.parser import Parser
from gherkin.token_scanner import TokenScanner

from .ast_node.gherkin_document import GherkinDocument
from .errors import DeserializeError, InvalidInput
from .utils import camel_to_snake_case, remove_trailing_spaces

T = TypeVar("T")


class CustomConverter(Converter):
    def structure_attrs_fromdict(self, obj: Dict[str, Any], cls: Type[T]) -> T:
        # Make sure the type in the parsed object matches the class we use
        # to structure the object
        if "type" in obj:
            type_name = obj.pop("type")
            cls_name = cls.__name__
            assert type_name == cls_name, f"{type_name} does not match {cls_name}"

        # Note that keys are in camelCase convention, for example, tableHeader,
        # tableBody. Therefore, we need to convert the keys to snake_case.
        transformed_obj = {}
        for key, value in obj.items():
            if isinstance(value, str):
                # For some types of node, the indentation of the lines is included
                # in the value of such nodes. Then the indentation can be changed after
                # formatting. Therefore, we need to dedent the value here for consistent
                # results. We also need to remove trailing spaces.
                value = remove_trailing_spaces(value)
                value = textwrap.dedent(value)

            transformed_obj[camel_to_snake_case(key)] = value

        return super(CustomConverter, self).structure_attrs_fromdict(
            transformed_obj, cls
        )


converter = CustomConverter()


# noinspection PyMissingConstructor
class StringOnlyTokenScanner(TokenScanner):
    """
    A replacement for Gherkin's TokenScanner that doesn't load from files.

    This is necessary to prevent "path too long for Windows" errors when Windows
    treats large feature files as paths on the file system (bug #34).
    """

    def __init__(self, content):
        self.io = io.StringIO(content)
        self.line_number = 0


def parse(content: str) -> GherkinDocument:
    """
    Parse the content of a file to an AST.
    """
    parser = Parser()

    try:
        parse_result = parser.parse(StringOnlyTokenScanner(content))
    except ParserError as e:
        raise InvalidInput(e) from e

    try:
        result = converter.structure(parse_result, GherkinDocument)
    except Exception as e:
        raise DeserializeError(f"{type(e).__name__}: {e}") from e

    return result



from typing import Optional, Tuple

from ._base import prepare
from .comment import Comment
from .feature import Feature


@prepare
class GherkinDocument:
    comments: Tuple[Comment, ...]
    feature: Optional[Feature] = None

    def __iter__(self):
        yield from self.comments

        if self.feature is not None:
            yield from self.feature



class BaseError(Exception):
    pass


class InvalidInput(BaseError):
    """
    Raised when the input file cannot be parsed.
    """


class DeserializeError(BaseError):
    """
    Raised when the parse result cannot be deserialized to an AST.
    """


class InternalError(BaseError):
    """
    Raised when something happens anomaly in the process of reformatting.
    """


class EquivalentError(InternalError):
    """
    Raised when the reformatted document is not equivalent to the original one.
    """


class StableError(InternalError):
    """
    Raised when we obtain a different document after reformatting the second time.
    """


class EmptySources(BaseError):
    """
    Raised when there is no file to reformat.
    """


class BaseWarning(Warning):
    pass


class MissingExamplesWarning(BaseWarning):
    """
    Raised when examples are missing in ScenarioOutline
    """


class EmptyExamplesWarning(BaseWarning):
    """
    Raised when an examples table is empty
    """


class NothingChanged(BaseWarning):
    """
    Raised when reformatted code is the same as source.
    """



from typing import Tuple

from ._base import prepare
from .location import LocationMixin
from .table_row import TableRow


@prepare
class DataTable(LocationMixin):
    rows: Tuple[TableRow, ...]

    def __iter__(self):
        yield self

        yield from self.rows



import difflib
import io
import re
import tempfile
import tokenize
from contextlib import nullcontext
from functools import lru_cache, partial
from pathlib import Path
from typing import IO, AnyStr, BinaryIO, Tuple, Union

import click
from wcwidth import wcswidth

out = partial(click.secho, bold=True, err=True)
err = partial(click.secho, fg="red", err=True)

_first_cap_re = re.compile(r"(.)([A-Z][a-z]+)")
_all_cap_re = re.compile(r"([a-z0-9])([A-Z])")


@lru_cache()
def camel_to_snake_case(name: str) -> str:
    """
    Convert camelCase to snake_case.
    Taken from https://stackoverflow.com/a/1176023/2585762.
    """
    s1 = _first_cap_re.sub(r"\1_\2", name)
    return _all_cap_re.sub(r"\1_\2", s1).lower()


def dump_to_file(*output: str) -> str:
    """
    Dump `output` to a temporary file. Return path to the file.
    """
    with tempfile.NamedTemporaryFile(
        mode="w", prefix="rfmt-ghk_", suffix=".log", delete=False, encoding="utf-8"
    ) as f:
        for lines in output:
            f.write(lines)
            if lines and lines[-1] != "\n":
                f.write("\n")
    return f.name


def diff(a: str, b: str, a_name: str, b_name: str) -> str:
    """Return a unified diff string between strings `a` and `b`."""
    a_lines = [line + "\n" for line in a.split("\n")]
    b_lines = [line + "\n" for line in b.split("\n")]
    return "".join(
        difflib.unified_diff(a_lines, b_lines, fromfile=a_name, tofile=b_name, n=5)
    )


_beginning_spaces_re = re.compile(r"^(\s*).*")


def extract_beginning_spaces(string: str) -> str:
    return _beginning_spaces_re.findall(string)[0]


def remove_trailing_spaces(string: str) -> str:
    lines = string.splitlines()
    return "\n".join(line.rstrip() for line in lines)


def decode_stream(src: BinaryIO) -> Tuple[str, str, str]:
    """
    Return a tuple of (decoded_contents, encoding, newline).

    `newline` is either CRLF or LF but `decoded_contents` is decoded with
    universal newlines (i.e. only contains LF).
    """
    # in case the source is not seekable, read into memory now
    src = io.BytesIO(src.read())
    encoding, lines = tokenize.detect_encoding(src.readline)
    if not lines:
        return "", encoding, "\n"

    newline = "\r\n" if b"\r\n" == lines[0][-2:] else "\n"
    src.seek(0)
    with io.TextIOWrapper(src, encoding) as tiow:
        return tiow.read(), encoding, newline


@lru_cache()
def get_display_width(text: str) -> int:
    """
    Get the display width of a string.

    When displayed in a console, some characters can have a width of 2 (for
    example, East Asian characters) or a width of 0 (for example, combining
    diacritic characters). Use the wcwidth package to find the correct display
    width for strings including these characters.

    Some characters do not have a width (for example, control characters like
    the bell character). If a string contains any such characters, then
    wcwidth.wcswidth returns -1. In this case, use the number of code points as
    a fallback.
    """
    width = wcswidth(text)
    if width < 0:
        width = len(text)
    return width


def open_stream_or_path(stream_or_path: Union[IO[AnyStr], Path], mode: str):
    if isinstance(stream_or_path, Path):
        return open(stream_or_path, mode)
    else:
        return nullcontext(stream_or_path)



from typing import Union

from .background import Background
from .comment import Comment
from .data_table import DataTable
from .doc_string import DocString
from .examples import Examples
from .feature import Feature
from .gherkin_document import GherkinDocument
from .location import Location
from .scenario import Scenario
from .scenario_outline import ScenarioOutline
from .step import Step
from .table_cell import TableCell
from .table_row import TableRow
from .tag import Tag
from .tag_group import TagGroup

Node = Union[
    Background,
    Comment,
    DataTable,
    DocString,
    Examples,
    Feature,
    GherkinDocument,
    Location,
    Scenario,
    ScenarioOutline,
    Step,
    TableCell,
    TableRow,
    Tag,
    TagGroup,
]



from itertools import chain
from typing import Optional, Tuple, Union

from ._base import prepare
from .background import Background
from .location import LocationMixin
from .scenario import Scenario
from .scenario_outline import ScenarioOutline
from .tag import Tag


@prepare
class Feature(LocationMixin):
    language: str
    keyword: str
    name: str
    children: Tuple[Union[Background, Scenario, ScenarioOutline], ...]
    tags: Tuple[Tag, ...]
    description: Optional[str] = None

    def __iter__(self):
        yield from self.tags

        yield self

        yield from chain.from_iterable(self.children)



from typing import Optional, Union

from ._base import prepare
from .data_table import DataTable
from .doc_string import DocString
from .location import LocationMixin


@prepare
class Step(LocationMixin):
    keyword: str
    text: str
    argument: Optional[Union[DataTable, DocString]] = None

    def __iter__(self):
        yield self

        if self.argument is not None:
            yield from self.argument



from attr import dataclass


def prepare(cls=None, slots=True, frozen=True, eq=False):
    """
    A common class decorator to decorate AST node classes. We can either use `@prepare`
    with default parameters, or `@prepare(...)` to override the default values of the
    parameters. By default, `eq=False` makes the objects hashable, and the hash is an
    object's id. Therefore, every AST node is unique, even if they have identical
    attributes (think of two identical rows or steps at two different places in a
    document).
    """
    wrapper = dataclass(slots=slots, frozen=frozen, eq=eq)

    if cls is None:
        return wrapper

    return wrapper(cls)



from typing import Tuple, Union

from ._base import prepare
from .examples import Examples
from .feature import Feature
from .location import LocationMixin
from .scenario import Scenario
from .scenario_outline import ScenarioOutline
from .tag import Tag


@prepare
class TagGroup(LocationMixin):
    members: Tuple[Tag, ...]
    context: Union[Examples, Feature, Scenario, ScenarioOutline]



from typing import Tuple

from ._base import prepare
from .location import LocationMixin
from .table_cell import TableCell


@prepare
class TableRow(LocationMixin):
    cells: Tuple[TableCell, ...]

    def __len__(self):
        return len(self.cells)

    def __getitem__(self, item):
        return self.cells[item]



from attr import attrib

from ._base import prepare
from .location import LocationMixin


def escape_doc_string_value(text: str) -> str:
    """Escape triple-quotes in doc strings."""
    return text.replace('"""', '\\"\\"\\"')


@prepare
class DocString(LocationMixin):
    content: str = attrib(converter=escape_doc_string_value)

    def __iter__(self):
        yield self



from attr import attrib

from ._base import prepare


@prepare(eq=True)
class Location:
    line: int
    column: int


@prepare
class LocationMixin:
    location: Location = attrib(eq=False, repr=False)



import sys
import traceback
from io import TextIOWrapper
from pathlib import Path
from typing import BinaryIO, Iterable, Iterator, Set, Tuple, Union

from .ast_node import GherkinDocument
from .errors import (
    BaseError,
    EmptySources,
    EquivalentError,
    InternalError,
    NothingChanged,
    StableError,
)
from .formatter import LineGenerator
from .options import NewlineMode, Options, WriteBackMode
from .parser import parse
from .report import Report
from .utils import decode_stream, diff, dump_to_file, err, open_stream_or_path

REPORT_URL = "https://github.com/ducminh-phan/reformat-gherkin/issues"

NEWLINE_FROM_OPTION = {NewlineMode.CRLF: "\r\n", NewlineMode.LF: "\n"}


def find_sources(src: Iterable[str]) -> Set[Path]:
    sources: Set[Path] = set()

    for s in src:
        path = Path(s).resolve()
        if path.is_dir():
            sources.update(path.rglob("*.feature"))
        elif path.is_file():
            # If a file was explicitly given, we don't care about its extension
            sources.add(path)
        else:  # pragma: no cover
            err(f"Invalid path: {s}")

    return sources


def reformat(src: Tuple[str], report: Report, *, options: Options):
    use_stdin = "-" in src
    sources = find_sources(filter((lambda it: it != "-"), src))

    if not sources and not use_stdin:
        raise EmptySources

    if use_stdin:
        changed = reformat_stdin(options=options)
        report.done("stdin", changed)

    for path in sources:
        try:
            changed = reformat_single_file(path, options=options)
            report.done(str(path), changed)
        except Exception as e:
            report.failed(path, str(e))


def reformat_stdin(*, options: Options) -> bool:
    output = sys.stdout.buffer if options.write_back == WriteBackMode.INPLACE else None
    return reformat_stream_or_path(
        sys.stdin.buffer, output, force_write=True, options=options
    )


def reformat_single_file(path: Path, *, options: Options) -> bool:
    out_path = path if options.write_back == WriteBackMode.INPLACE else None
    return reformat_stream_or_path(path, out_path, options=options)


def reformat_stream_or_path(
    in_stream_or_path: Union[BinaryIO, Path],
    out_stream_or_path: Union[None, BinaryIO, Path],
    *,
    force_write: bool = False,
    options: Options,
) -> bool:
    with open_stream_or_path(in_stream_or_path, "rb") as in_stream:
        src_contents, encoding, existing_newline = decode_stream(in_stream)

    newline = NEWLINE_FROM_OPTION.get(options.newline, existing_newline)
    newline_changed = newline != existing_newline

    content_changed = True
    try:
        dst_contents = format_file_contents(src_contents, options=options)
    except NothingChanged:
        content_changed = False
        dst_contents = src_contents

    will_write = force_write or content_changed or newline_changed

    if will_write and out_stream_or_path is not None:
        with open_stream_or_path(out_stream_or_path, "wb") as out_stream:
            tiow = TextIOWrapper(out_stream, encoding=encoding, newline=newline)
            tiow.write(dst_contents)
            # Ensures that the underlying stream is not closed when the
            # TextIOWrapper is garbage collected. We don't want to close a
            # stream that was passed to us.
            tiow.detach()

    return content_changed or newline_changed


def format_file_contents(src_contents: str, *, options: Options) -> str:
    """
    Reformat the contents of a file and return new contents. Raise NothingChanged
    if the contents were not changed after reformatting.

    If `options.fast` is False, additionally confirm that the reformatted file is
    valid by calling :func:`assert_equivalent` and :func:`assert_stable` on it.
    """
    if src_contents.strip() == "":
        raise NothingChanged

    dst_contents = format_str(src_contents, options=options)
    if src_contents == dst_contents:
        raise NothingChanged

    if not options.fast:
        assert_equivalent(src_contents, dst_contents)
        assert_stable(src_contents, dst_contents, options=options)

    return dst_contents


def format_str(src_contents: str, *, options: Options) -> str:
    """
    Reformat a string and return new contents.
    """
    ast = parse(src_contents)

    line_generator = LineGenerator(
        ast, options.step_keyword_alignment, options.tag_line_mode, options.indent
    )
    lines = line_generator.generate()

    return "\n".join(lines)


def assert_equivalent(src: str, dst: str) -> None:
    """
    Raise EquivalentError if `src` and `dst` aren't equivalent.
    """

    def _v(ast: GherkinDocument) -> Iterator[str]:
        """
        Simple visitor generating strings to compare ASTs by content
        """
        for node in ast:
            yield repr(node)

    src_ast = parse(src)

    try:
        dst_ast = parse(dst)
    except BaseError as exc:
        log = dump_to_file("".join(traceback.format_tb(exc.__traceback__)), dst)
        raise InternalError(
            f"INTERNAL ERROR: Invalid file contents are produced:\n"
            f"{exc}\n"
            f"Please report a bug on {REPORT_URL}.\n"
            f"This invalid output might be helpful:\n"
            f"{log}\n"
        ) from exc

    src_ast_str = "\n".join(_v(src_ast))
    dst_ast_str = "\n".join(_v(dst_ast))

    if src_ast_str != dst_ast_str:
        log = dump_to_file(diff(src_ast_str, dst_ast_str, "src", "dst"))
        raise EquivalentError(
            f"INTERNAL ERROR: The new content produced is not equivalent to "
            f"the source.\n"
            f"Please report a bug on {REPORT_URL}.\n"
            f"This diff might be helpful: {log}\n"
        )


def assert_stable(src: str, dst: str, *, options: Options) -> None:
    """
    Raise StableError if `dst` reformats differently the second time.
    """
    new_dst = format_str(dst, options=options)
    if dst != new_dst:
        log = dump_to_file(
            diff(src, dst, "source", "first pass"),
            diff(dst, new_dst, "first pass", "second pass"),
        )
        raise StableError(
            f"INTERNAL ERROR: Different contents are produced on the second pass "
            f"of the formatter.\n"
            f"Please report a bug on {REPORT_URL}.\n"
            f"This diff might be helpful: {log}\n"
        ) from None



from typing import Optional, Tuple

from ._base import prepare
from .location import LocationMixin
from .table_row import TableRow
from .tag import Tag


@prepare
class Examples(LocationMixin):
    keyword: str
    name: str
    tags: Tuple[Tag, ...]
    table_header: Optional[TableRow] = None
    table_body: Optional[Tuple[TableRow, ...]] = None
    description: Optional[str] = None

    def __iter__(self):
        yield from self.tags

        yield self

        if self.table_header is not None:
            yield self.table_header

        if self.table_body is not None:
            yield from self.table_body



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['reformat_gherkin', 'reformat_gherkin.ast_node']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=20.2,<21.0',
 'cattrs>=1.1.0,<2.0.0',
 'click>=7.0,<8.0',
 'gherkin-official==4.1.3',
 'pyyaml>=5.1,<6.0',
 'wcwidth>=0.1.7,<0.2.0']

entry_points = \
{'console_scripts': ['reformat-gherkin = reformat_gherkin.cli:main']}

setup_kwargs = {
    'name': 'reformat-gherkin',
    'version': '2.1.0',
    'description': 'Formatter for Gherkin language',
    'long_description': "# Reformat-gherkin\n\n[![Build Status](https://dev.azure.com/alephvn/reformat-gherkin/_apis/build/status/ducminh-phan.reformat-gherkin?branchName=master)](https://dev.azure.com/alephvn/reformat-gherkin/_build/latest?definitionId=1&branchName=master)\n&nbsp; [![Build Status](https://travis-ci.com/ducminh-phan/reformat-gherkin.svg?branch=master)](https://travis-ci.com/ducminh-phan/reformat-gherkin)\n&nbsp; [![Coverage Status](https://coveralls.io/repos/github/ducminh-phan/reformat-gherkin/badge.svg?branch=master)](https://coveralls.io/github/ducminh-phan/reformat-gherkin?branch=master)\n\n[![Maintainability](https://api.codeclimate.com/v1/badges/16718a231901c293215d/maintainability)](https://codeclimate.com/github/ducminh-phan/reformat-gherkin/maintainability)\n&nbsp; [![Codacy Badge](https://api.codacy.com/project/badge/Grade/e675ca51b6ac436a980facbcf04b8e5a)](https://www.codacy.com/app/ducminh-phan/reformat-gherkin)\n\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n&nbsp; [![PyPI](https://img.shields.io/pypi/v/reformat-gherkin.svg)](https://pypi.org/project/reformat-gherkin/)\n&nbsp; [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/python/black)\n\n## Table of Contents\n\n- [About](#about)\n- [Installation](#installation)\n- [Usage](#usage)\n- [Version control integration](#version-control-integration)\n- [Acknowledgements](#acknowledgements)\n\n## About\n\nReformat-gherkin automatically formats Gherkin files. It ensures a consistent\nlook regardless of who wrote the file.\n\nIt can be used either as a command-line tool, or a\n[pre-commit](https://pre-commit.com/) hook.\n\n## Installation\n\nInstall reformat-gherkin using [pip](https://pypi.org/project/pip/).\n\n```bash\npip install reformat-gherkin\n```\n\nIt requires [Python 3.7+](https://www.python.org/downloads/) to run.\n\n## Usage\n\nTo get started straight away:\n\n```bash\nreformat-gherkin {source_file_or_directory}\n```\n\nYou should get good results without specifying any options, as reformat-gherkin\nuses sensible defaults.\n\n### Command-line options\n\nYou can list the available options by running `reformat-gherkin --help`.\n\n```text\nUsage: reformat-gherkin [OPTIONS] [SRC]...\n\n  Reformat the given SRC files and all .feature files in SRC folders. If -\n  is passed as a file, reformat stdin and print the result to stdout.\n\nOptions:\n  --check                         Don't write the files back, just return the\n                                  status. Return code 0 means nothing would\n                                  change. Return code 1 means some files would\n                                  be reformatted. Return code 123 means there\n                                  was an internal error.\n\n  -a, --alignment [left|right]    Specify the alignment of step keywords\n                                  (Given, When, Then,...). If specified, all\n                                  statements after step keywords are left-\n                                  aligned, spaces are inserted before/after\n                                  the keywords to right/left align them. By\n                                  default, step keywords are left-aligned, and\n                                  there is a single space between the step\n                                  keyword and the statement.\n\n  -n, --newline [LF|CRLF]         Specify the line separators when formatting\n                                  files inplace. If not specified, line\n                                  separators are preserved.\n\n  --fast / --safe                 If --fast given, skip the sanity checks of\n                                  file contents. [default: --safe]\n\n  --single-line-tags / --multi-line-tags\n                                  If --single-line-tags given, output\n                                  consecutive tags on one line. If --multi-\n                                  line-tags given, output one tag per line.\n                                  [default: --single-line-tags]\n\n  --tab-width INTEGER             Specify the number of spaces per\n                                  indentation-level. [default: 2]\n\n  --use-tabs                      Indent lines with tabs instead of spaces.\n  --config FILE                   Read configuration from FILE.\n  --version                       Show the version and exit.\n  --help                          Show this message and exit.\n```\n\nReformat-gherkin is a well-behaved Unix-style command-line tool:\n\n- it does nothing if no sources are passed to it;\n- it will read from standard input and write to standard output if - is used as the filename;\n- it only outputs messages to users on standard error;\n- it exits with code 0 unless an internal error occurred (or --check was used).\n\n### Config file\n\nReformat-gherkin can read project-specific default values for its command line\noptions from a `.reformat-gherkin.yaml` file.\n\nBy default, `reformat-gherkin` looks for the config file starting from the\ncommon base directory of all files and directories passed on the command line.\nIf it's not there, it looks in parent directories. It stops looking when it\nfinds the file, or a .git directory, or a .hg directory, or the root of the file\nsystem, whichever comes first.\n\nExample config file:\n\n```yaml\ncheck: False\nalignment: left\ntab_width: 4\n```\n\n## Version control integration\n\nYou can integrate reformat-gherkin into your version control workflow by using\n[pre-commit](https://pre-commit.com/). Once you have installed pre-commit, add\nthis to the `.pre-commit-config.yaml` file in your repository:\n\n```text\nrepos:\n  - repo: https://github.com/ducminh-phan/reformat-gherkin\n    rev: stable\n    hooks:\n      - id: reformat-gherkin\n```\n\nThen run `pre-commit install` and you're ready to go.\n\n## Acknowledgements\n\nThis project is inspired by [black](https://github.com/psf/black). Some\nfunctions are taken from `black`'s source code.\n",
    'author': 'Duc-Minh Phan',
    'author_email': 'alephvn@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ducminh-phan/reformat-gherkin',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import Optional, Tuple

import click

from .config import read_config_file
from .core import reformat
from .errors import EmptySources
from .options import (
    AlignmentMode,
    NewlineMode,
    Options,
    TagLineMode,
    WriteBackMode,
    get_indent_from_configuration,
)
from .report import Report
from .utils import out
from .version import __version__


@click.command()
@click.argument(
    "src",
    nargs=-1,
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=True,
        readable=True,
        resolve_path=True,
        allow_dash=True,
    ),
    is_eager=True,
)
@click.option(
    "--check",
    is_flag=True,
    help=(
        "Don't write the files back, just return the status. Return code 0 "
        "means nothing would change. Return code 1 means some files would be "
        "reformatted. Return code 123 means there was an internal error."
    ),
)
@click.option(
    "-a",
    "--alignment",
    type=click.Choice([AlignmentMode.LEFT.value, AlignmentMode.RIGHT.value]),
    help=(
        "Specify the alignment of step keywords (Given, When, Then,...). "
        "If specified, all statements after step keywords are left-aligned, "
        "spaces are inserted before/after the keywords to right/left align them. "
        "By default, step keywords are left-aligned, and there is a single "
        "space between the step keyword and the statement."
    ),
)
@click.option(
    "-n",
    "--newline",
    type=click.Choice([NewlineMode.LF.value, NewlineMode.CRLF.value]),
    help=(
        "Specify the line separators when formatting files inplace. "
        "If not specified, line separators are preserved."
    ),
)
@click.option(
    "--fast/--safe",
    is_flag=True,
    help="If --fast given, skip the sanity checks of file contents. [default: --safe]",
)
@click.option(
    "--single-line-tags/--multi-line-tags",
    is_flag=True,
    default=True,
    help=(
        "If --single-line-tags given, output consecutive tags on one line. "
        "If --multi-line-tags given, output one tag per line. "
        "[default: --single-line-tags]"
    ),
)
@click.option(
    "--tab-width",
    type=int,
    default=2,
    help="Specify the number of spaces per indentation-level. [default: 2]",
)
@click.option(
    "--use-tabs",
    is_flag=True,
    default=False,
    help="Indent lines with tabs instead of spaces.",
)
@click.option(
    "--config",
    type=click.Path(
        exists=True, file_okay=True, dir_okay=False, readable=True, allow_dash=False
    ),
    is_eager=True,
    callback=read_config_file,
    help="Read configuration from FILE.",
)
@click.version_option(version=__version__)
@click.pass_context
def main(
    ctx: click.Context,
    src: Tuple[str],
    check: bool,
    alignment: Optional[str],
    newline: Optional[str],
    fast: bool,
    single_line_tags: bool,
    tab_width: int,
    use_tabs: bool,
    config: Optional[str],
) -> None:
    """
    Reformat the given SRC files and all .feature files in SRC folders. If -
    is passed as a file, reformat stdin and print the result to stdout.
    """
    if config:
        out(f"Using configuration from {config}.", bold=False, fg="blue")

    write_back_mode = WriteBackMode.from_configuration(check)
    alignment_mode = AlignmentMode.from_configuration(alignment)
    newline_mode = NewlineMode.from_configuration(newline)
    tag_line_mode = TagLineMode.from_configuration(single_line_tags)
    indent = get_indent_from_configuration(tab_width, use_tabs)

    options = Options(
        write_back=write_back_mode,
        step_keyword_alignment=alignment_mode,
        newline=newline_mode,
        fast=fast,
        tag_line_mode=tag_line_mode,
        indent=indent,
    )

    report = Report(check=check)
    try:
        reformat(src, report, options=options)
    except EmptySources:
        out("No paths given. Nothing to do 😴")
        ctx.exit(0)

    bang = "💥 💔 💥" if report.return_code else "✨ 🍰 ✨"
    out(f"All done! {bang}")
    click.secho(str(report), err=True)
    ctx.exit(report.return_code)
